## Copyright (C) 2009-2012 Eric Chassande-Mottin, CNRS (France)
## Parts Copyright (C) 2012 Philip Nienhuis
##
## This file is part of Octave.
##
## Octave is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or (at
## your option) any later version.
##
## Octave is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with Octave; see the file COPYING.  If not, see
## <http://www.gnu.org/licenses/>.

## -*- texinfo -*-
## @deftypefn  {Function File} {[@var{a}, @dots{}] =} strread (@var{str})
## @deftypefnx {Function File} {[@var{a}, @dots{}] =} strread (@var{str}, @var{format})
## @deftypefnx {Function File} {[@var{a}, @dots{}] =} strread (@var{str}, @var{format}, @var{format_repeat})
## @deftypefnx {Function File} {[@var{a}, @dots{}] =} strread (@var{str}, @var{format}, @var{prop1}, @var{value1}, @dots{})
## @deftypefnx {Function File} {[@var{a}, @dots{}] =} strread (@var{str}, @var{format}, @var{format_repeat}, @var{prop1}, @var{value1}, @dots{})
## Read data from a string.
##
## The string @var{str} is split into words that are repeatedly matched to the
## specifiers in @var{format}.  The first word is matched to the first
## specifier, the second to the second specifier and so forth.  If there are
## more words than specifiers, the process is repeated until all words have
## been processed.
##
## The string @var{format} describes how the words in @var{str} should be
## parsed.
## It may contain any combination of the following specifiers:
##
## @table @code
## @item %s
## The word is parsed as a string.
##
## @itemx %f
## @itemx %n
## The word is parsed as a number and converted to double.
##
## @item  %d
## @itemx %u
## The word is parsed as a number and converted to int32.
##
## @item %q
## Read a text string enclosed in double quotes. NOTE: this is experimental
## and may not work as expected if the matching data text strings contain
## whitespace and/or delimiter characters. Multiple consecutive whitespaces
## will be folded into one space in the returned strings.
##
## @item %*d, %*n, %*u, %*f, %*s, %*q
## The word is skipped.
##
## For %s and %d, %f, %n, %u and the associated %*s @dots{} specifiers an
## optional character width can be specified as %Ns, etc. where N is an
## integer > 1. For %f, format specifiers like %N.Mf are allowed.
##
## For %f, %u, and %d specifiers, a bit width can be specified and the
## relating read numbers will be casted to that bit width: %f32 to single,
## %f64 to double, %d8, %d16, %d32 and %d64 to int8, int16, int32, and int64,
## respectively; same for %u8 ... %u64 (uint8 ... uint64).
##
## @item %[ ]
## Pattern matching: read only characters matching those enclosed between the
## brackets til the first non-matching character or delimiter.
## @item %[^ ]
## Read only characters NOT matching those between %[^ and ] til the first
## matching character.
##
## @item %*[ ], %*[^ ]
## Skip matching or non-matching characters, respectively.
##
## If ']' should be part of the pattern it must be the first character of the
## actual pattern after the starting [ or [^: %[] ] or %[^] ]
##
## NOTE: patterns containing whitespace may not work as expected due to
## the way strread currently works. Patterns containing delimiters may not
## work at all, especially if the number of enclosed delimiters differs from 
## line to line. endofline (EOL) characters have a better chance to work.
##
## @item literals
## In addition the format may contain literal character strings; these will be
## skipped during reading.
## @end table
##
## Parsed word corresponding to the first specifier are returned in the first
## output argument and likewise for the rest of the specifiers.
##
## By default, @var{format} is @t{"%f"}, meaning that numbers are read from
## @var{str}.  This will do if @var{str} contains only numeric fields.
##
## For example, the string
##
## @example
## @group
## @var{str} = "\
## Bunny Bugs   5.5\n\
## Duck Daffy  -7.5e-5\n\
## Penguin Tux   6"
## @end group
## @end example
##
## @noindent
## can be read using
##
## @example
## [@var{a}, @var{b}, @var{c}] = strread (@var{str}, "%s %s %f");
## @end example
##
## Optional numeric argument @var{format_repeat} can be used for
## limiting the number of items read:
##
## @table @asis
## @item -1
## (default) read all of the string until the end.
##
## @item N
## Read N times @var{nargout} items.  0 (zero) is an acceptable
## value for @var{format_repeat}.
## @end table
##
## The behavior of @code{strread} can be changed via property-value
## pairs.  The following properties are recognized:
##
## @table @asis
## @item "commentstyle"
## Parts of @var{str} are considered comments and will be skipped.
## @var{value} is the comment style and can be any of the following.
##
## @itemize
## @item "shell"
## Everything from @code{#} characters to the nearest end-of-line is skipped.
##
## @item "c"
## Everything between @code{/*} and @code{*/} is skipped.
##
## @item "c++"
## Everything from @code{//} characters to the nearest end-of-line is skipped.
##
## @item "matlab"
## Everything from @code{%} characters to the nearest end-of-line is skipped.
##
## @item user-supplied.  Two options:
## (1) One string, or 1x1 cell string: Skip everything to the right of it;
## (2) 2x1 cell string array: Everything between the left and right strings
## is skipped.
## @end itemize
##
## @item "delimiter"
## Any character in @var{value} will be used to split @var{str} into words
## (default value = any whitespace).
##
## @item "emptyvalue":
## Value to return for empty numeric values in non-whitespace delimited data.
## The default is NaN@.  When the data type does not support NaN
## (int32 for example), then default is zero.
##
## @item "multipledelimsasone"
## Treat a series of consecutive delimiters, without whitespace in between,
## as a single delimiter.  Consecutive delimiter series need not be vertically
## "aligned".
##
## @item "treatasempty"
## Treat single occurrences (surrounded by delimiters or whitespace) of the
## string(s) in @var{value} as missing values.
##
## @item "returnonerror"
## If @var{value} true (1, default), ignore read errors and return normally.
## If false (0), return an error.
##
## @item "whitespace"
## Any character in @var{value} will be interpreted as whitespace and
## trimmed; the string defining whitespace must be enclosed in double
## quotes for proper processing of special characters like \t.
## The default value for whitespace = " \b\r\n\t" (note the space).
## Unless whitespace is set to '' (empty) AND at least one "%s" format
## conversion specifier is supplied, a space is always part of whitespace.
##
## @end table
##
## When the number of words in @var{str} doesn't match an exact multiple
## of the number of format conversion specifiers, strread's behavior
## depends on the last character of @var{str}:
##
## @table @asis
## @item last character = "\n"
## Data columns are padded with empty fields or Nan so that all columns
## have equal length 
##
## @item last character is not "\n"
## Data columns are not padded; strread returns columns of unequal length
##
## @end table
##
## @seealso{textscan, textread, load, dlmread, fscanf}
## @end deftypefn

function varargout = strread (str, format = "%f", varargin)

  ## Check input
  if (nargin < 1)
    print_usage ();
  endif
  
  if (isempty (str))
    ## Return empty args (no match), rather than raising an error
    varargout = cell (1, nargout);
    return;
  endif

  if (isempty (format))
    format = "%f";
  endif

  if (! ischar (str) || ! ischar (format))
    error ("strread: STR and FORMAT arguments must be strings");
  endif

  ## Parse format string to compare number of conversion fields and nargout
  nfields = length (strfind (format, "%")) - length (strfind (format, "%*"));
  ## If str only has numeric fields, a (default) format ("%f") will do.
  ## Otherwise:
  if ((max (nargout, 1) != nfields) && ! strcmp (format, "%f"))
    error ("strread: the number of output variables must match that specified by FORMAT");
  endif

  ## Check for format string repeat count
  format_repeat_count = -1;
  if (nargin > 2 && isnumeric (varargin{1}))
    if (varargin{1} >= 0)
      format_repeat_count = varargin{1};
    endif
    if (nargin > 3)
      varargin = varargin(2:end);
    else
      varargin = {};
    endif
  endif

  ## Parse options.  First initialize defaults
  comment_flag = false;
  delimiter_str = "";
  empty_str = "";
  eol_char = "";
  err_action = 0;
  mult_dlms_s1 = false;
  numeric_fill_value = NaN;
  white_spaces = " \b\r\n\t";
  for n = 1:2:length (varargin)
    switch (lower (varargin{n}))
      case "bufsize"
        ## We could synthesize this, but that just seems weird...
        warning ("strread: property 'bufsize' is not implemented");
      case "commentstyle"
        comment_flag = true;
        switch (lower (varargin{n+1}))
          case "c"
            [comment_start, comment_end] = deal ("/*", "*/");
          case "c++"
            [comment_start, comment_end] = deal ("//", "eol_char");
          case "shell"
            [comment_start, comment_end] = deal ("#" , "eol_char");
          case "matlab"
            [comment_start, comment_end] = deal ("%" , "eol_char");
          otherwise
            if (ischar (varargin{n+1}) ||
               (numel (varargin{n+1}) == 1 && iscellstr (varargin{n+1})))
              [comment_start, comment_end] = deal (char (varargin{n+1}), "eol_char");
            elseif (iscellstr (varargin{n+1}) && numel (varargin{n+1}) == 2)
              [comment_start, comment_end] = deal (varargin{n+1}{:});
            else
              ## FIXME - a user may have numeric values specified: {'//', 7}
              ##         this will lead to an error in the warning message
              error ("strread: unknown or unrecognized comment style '%s'",
                      varargin{n+1});
            endif
        endswitch
      case "delimiter"
        delimiter_str = varargin{n+1};
        if (strcmp (typeinfo (delimiter_str), "sq_string"))
          delimiter_str = do_string_escapes (delimiter_str);
        endif
      case "emptyvalue"
        numeric_fill_value = varargin{n+1};
      case "expchars"
        warning ("strread: property 'expchars' is not implemented");
      case "whitespace"
        white_spaces = varargin{n+1};
        if (strcmp (typeinfo (white_spaces), "sq_string"))
          white_spaces = do_string_escapes (white_spaces);
        endif
      ## The following parameters are specific to textscan and textread
      case "endofline"
        eol_char = varargin{n+1};
        if (strcmp (typeinfo (eol_char), "sq_string"))
          eol_char = do_string_escapes (eol_char);
        endif
      case "returnonerror"
        err_action = varargin{n+1};
      case "multipledelimsasone"
        mult_dlms_s1 = varargin{n+1};
      case "treatasempty"
        if (iscellstr (varargin{n+1}))
          empty_str = varargin{n+1};
        elseif (ischar (varargin{n+1}))
          empty_str = varargin(n+1);
        else
          error ("strread: 'treatasempty' value must be string or cellstr");
        endif
      otherwise
        warning ("strread: unknown property '%s'", varargin{n});
    endswitch
  endfor

  ## First parse of FORMAT
  if (strcmpi (strtrim (format), "%f"))
    ## Default format specified.  Expand it (to desired nargout)
    fmt_words = cell (nargout, 1);
    fmt_words (1:nargout) = format;
    idw = zeros (1, numel (fmt_words));
  else
    ## Determine the number of words per line as a first guess.  Forms
    ## like %f<literal>) (w/o delimiter in between) are fixed further on
    format = strrep (format, "%", " %");
    fmt_words = regexp (format, '[^ ]+', "match");

    ## Find position of conversion specifiers (they start with %)
    idy2 = find (! cellfun ("isempty", regexp (fmt_words, '^%')));

    ## Check for unsupported format specifiers
    errpat = '^%\*?c';
    if (! all (cellfun ("isempty", regexp (fmt_words(idy2), errpat))))
      error ("strread: %c format specifiers are not supported. Try %1s instead");
    endif

    ## Check for whitespace or delimiters in %[] style fmt_words.
    ## After all, patterns like "%[] [%ab\ncdefg]" are allowed...
    ii = 1; num_words_per_line = numel (fmt_words);
    ## idw keeps track of which %[] fmt_words contain whitespace/delims
    idw = zeros (1, num_words_per_line);
    while ii <= num_words_per_line
      if (! isempty (regexp (fmt_words{ii}, '%\*?\[')))
        ## %[] style conv specifier. Find begin pos of actual pattern chars
        ## 1. Check if pattern characters contains ']' 
        es = regexp (fmt_words{ii}, '%(\*?\[\]|\*?\[\^\])', 'end');
        if (isempty (es))
          ## 2. No, assume pattern w/o ']'
          es = regexp (fmt_words{ii}, '%(\*?\[|\*?\[\^)', 'end');
        endif
        ## Check if closing ']' of pattern is in fmt_words{ii}.
        ## If fmt_words{i} ends after es we can skip searching ']'
        no_close_brckt = (es == length (fmt_words{ii}));
        if (no_close_brckt || isempty (strfind (fmt_words{ii}(es+1:end), ']')))
          ## Apparently one or more spaces / delimiters in pattern as fmt_words
          ## got split up there. Check next fmt_words where matching ] is
          jj = ii + 1; mtch_ptrn = 0;
          while jj <= num_words_per_line
            ## Even "%", "*", "[", "^" and "]" (in that order) could be in pattern
            if (! isempty (regexp (fmt_words{ii+1}, '%\*?\[\^')))
              if (mtch_ptrn)
                error ("strread - ambiguous format specifier # %d", ii);
              else
                mtch_ptrn = 1;
              endif
            endif
            if (isempty (regexp (fmt_words{jj}, '\]')))
              ## Found no matching ] yet, search next fmt_words
              ++jj;
            else 
              ## Found the missing ]. Concatenate adjacent fmt_words
              fmt_words{ii} = [fmt_words{ii:jj}];
              fmt_words(ii+1:jj) = [];
              num_words_per_line -= jj - ii;
              ## For %[^..], the delim/whitesp behavior is inverted
              idw(ii) = 1;
              idw(ii+1:jj) = [];
              jj = num_words_per_line + 1;
            endif
          endwhile
        else
          idw(ii) = ! isempty (regexp (fmt_words{ii}, '%\*?\[\^?'));
        endif
      endif
      ++ii;
    endwhile
    ## Just renew idy2
    idy2 = find (! cellfun ("isempty", regexp (fmt_words, '^%')));

    allwptrn = '([nfdu]8|[nfdu]16|[nfdu]32|[nfdu]64)';
    ## Format conversion specifiers following literals w/o space/delim
    ## in between are separate now.  Separate those w trailing literals
    a = strfind (fmt_words(idy2), "%");
    b = regexp (fmt_words(idy2), '(\[.*\]|[nfdusq])', 'end');
    for jj = 1:numel (a)
      ## From right to left to avoid losing track
      ii = numel (a) - jj + 1;
      [re] = regexp (fmt_words{idy2(ii)}, allwptrn, "end");
      if (!isempty (re))
        ## Add bit width specifier to conv. specifier
        b{ii}(1) = re;
      endif
      if (! (length (fmt_words{idy2(ii)}) == b{ii}(1)))
        ## Split fmt_words(ii) into % conv specifier and trailing literal
        fmt_words(idy2(ii)+1 : end+1) = fmt_words(idy2(ii) : end);
        idw(idy2(ii)+1 : end+1) = idw(idy2(ii) : end);
        idw(idy2(ii)+1) = 0;
        fmt_words{idy2(ii)} = fmt_words{idy2(ii)}(a{ii} : b{ii}(1));
        fmt_words{idy2(ii)+1} = fmt_words{idy2(ii)+1}(b{ii}+1:end);
      endif
    endfor
  endif
  num_words_per_line = numel (fmt_words);

  ## Special handling for CRLF EOL character in str
  if (! isempty (eol_char) && strcmp (eol_char, "\r\n"))
    ## Strip CR from CRLF sequences
    str = strrep (str, "\r\n", "\n");
    ## CR serves no further purpose in function
    eol_char = "\n";
  endif

  ## Remove comments in str
  if (comment_flag)
    ## Expand 'eol_char' here, after option processing which may have set value
    comment_end = regexprep (comment_end, "eol_char", eol_char);
    cstart = strfind (str, comment_start);
    cstop  = strfind (str, comment_end);
    ## Treat end of string as additional comment stop
    if (isempty (cstop) || cstop(end) != length (str))
      cstop(end+1) = length (str);
    endif
    if (! isempty (cstart))
      ## Ignore nested openers.
      [idx, cidx] = unique (lookup (cstop, cstart), "first");
      if (idx(end) == length (cstop))
        cidx(end) = []; # Drop the last one if orphaned.
      endif
      cstart = cstart(cidx);
    endif
    if (! isempty (cstop))
      ## Ignore nested closers.
      [idx, cidx] = unique (lookup (cstart, cstop), "first");
      if (idx(1) == 0)
        cidx(1) = []; # Drop the first one if orphaned.
      endif
      cstop = cstop(cidx);
    endif
    len = length (str);
    c2len = length (comment_end);
    str = cellslices (str, [1, cstop + c2len], [cstart - 1, len]);
    str = [str{:}];
  endif

  if (! isempty (white_spaces))
    ## For numeric fields, whitespace is always a delimiter, but not for text
    ## fields
    if (isempty (strfind (format, "%s")))
      ## Add whitespace to delimiter set
      delimiter_str = unique ([white_spaces delimiter_str]);
    else
      ## Remove any delimiter chars from white_spaces list
      white_spaces = setdiff (white_spaces, delimiter_str);
    endif
  endif
  if (isempty (delimiter_str))
    delimiter_str = " ";
  endif
  if (! isempty (eol_char))
    ## Add eol_char to delimiter collection
    delimiter_str = unique ([delimiter_str eol_char]);
    ## .. and remove it from whitespace collection
    white_spaces = strrep (white_spaces, eol_char, "");
  endif

  ## Check if pattern contains EOL (a special one as EOL can be used
  ## to reading until EOL while skipping other delimiters
  ideol = zeros (1, num_words_per_line);
  if (! isempty (eol_char))
    for ii=1:num_words_per_line
      if (idw(ii))
        ideol(ii) = ! isempty (strfind (fmt_words{ii}, eol_char));
      endif
    endfor
  endif
  if (sum(ideol))
    ## Prepend char(253) to all EOL in str so some trace of EOL is left
    ## FIXME this assumes char(253) doesn't occur in str
    str = strrep (str, eol_char, [char(253) eol_char]);
    ## Ensure last char was eol_char
    if (str(end) != char(253)); str(end+1) = char(253); endif
    fmt_words = strrep (fmt_words, eol_char, char(253));
  endif

  pad_out = 0;
  ## Trim whitespace if needed
  if (! isempty (white_spaces))
    ## Check if trailing "\n" might signal padding output arrays to equal size
    ## before it is trimmed away below
    if ((str(end) == 10) && (nargout > 1))
      pad_out = 1;
    endif
    ## Condense all repeated whitespace into one single space
    ## FIXME: this will also fold repeated whitespace in a char field
    rxp_wsp = sprintf ("[%s]+", white_spaces);
    str = regexprep (str, rxp_wsp, " ");
    ## Remove possible leading space at string
    if (str(1) == 32)
       str = str(2:end);
    endif
    ## Check for single delimiter followed/preceded by whitespace
    if (! isempty (delimiter_str))
      dlmstr = setdiff (delimiter_str, " ");
      rxp_dlmwsp = sprintf ("( [%s]|[%s] )", dlmstr, dlmstr);
      str = regexprep (str, rxp_dlmwsp, delimiter_str(1));
    endif
    ## Wipe leading and trailing whitespace on each line (it may be
    ## delimiter too)
    ## FIXME: Double strrep on str is enormously expensive of CPU time.
    ## Can this be eliminated
    if (! isempty (eol_char))
      str = strrep (str, [eol_char " "], eol_char);
      str = strrep (str, [" " eol_char], eol_char);
    endif
  endif

  ## Split 'str' into words
  words = split_by (str, delimiter_str, mult_dlms_s1, eol_char);
  if (! isempty (white_spaces))
    ## Trim leading and trailing 'white_spaces'. All whitespace has
    ## been converted to space above
    words = strtrim (words);
  endif
  num_words = numel (words);
  ## First guess at nr. of lines in file (ignoring leading/trailing literals)
  num_lines = ceil (num_words / num_words_per_line);

  ## Replace TreatAsEmpty char sequences by empty strings
  if (! isempty (empty_str))
    for ii = 1:numel (empty_str)
      ## "words" matching empty_str may still be bordered by literals
      idz = strncmp (empty_str{ii}, words, length (empty_str{ii}));
      words(idz) = {""};
    endfor
  endif

  ## fmt_words has been split properly now, but words{} has only been split on
  ## delimiter positions. As numeric fields can also be separated by
  ## whitespace, more splits may be needed.
  ## In addition, words may have been split within %[ ] matches if the matched
  ## strings contain delimiters. These splits must be undone.
  ## Also we don't know the number of lines (as EndOfLine may have been set to
  ## "" (empty) by the caller).
  ##
  ## We also may have to cope with 3 cases as far as literals go:
  ## A: Trailing literals (%f<literal>) w/o delimiter in between.
  ## B: Leading literals (<literal>%f) w/o delimiter in between.
  ## C. Skipping leftover parts of specified skip fields (%*N )
  ## Some words columns may have to be split further to fix these.
  ## To find out, we'll match fmt_words to the words array to see what
  ## needs to be done. fwptr tracks which {fmt_words}# starts in what {words}#

  ## Find indices and pointers to possible literals in fmt_words
  idf = cellfun ("isempty", strfind (fmt_words, "%"));
  ## Find indices and pointers to conversion specifiers with fixed width
  idg = ! cellfun ("isempty", regexp (fmt_words, '%\*?\d'));
  ## Find indices and pointers to regexp-like conversion specifiers
  idr = ! cellfun ("isempty", regexp (fmt_words, '%\*?\[.*\]', 'end'));
  ## Find %q (double-space) conversion specifiers
  idq = ! cellfun ("isempty", regexp (fmt_words, '^%\*?q'));
  
  idy = find (idf | idg | idr | idq);

  ## Find indices to numeric conversion specifiers
  idn = ! cellfun ("isempty", regexp (fmt_words, '%[dnfu]'));

  ## If needed, split up columns in three steps:
  if (! isempty (idy))
    ## Try-catch because complexity of strings to read can be infinite
    try

      ## 1. Assess "period" in the split-up words array ( < num_words_per_line).
      ## Could be done using EndOfLine but that prohibits EndOfLine = "" option.
      ## Alternative below goes by simply parsing a first grab of words
      ## and matching fmt_words to words until the fmt_words array is exhausted.
      ## iwrd: ptr to current analyzed word; iwrdp: ptr to pos before analyzed char
      iwrd = 1; iwrdp = 0; iwrdl = length (words{1});
      sttc = endc = fwptr = zeros (1, numel (fmt_words));
      ii = 1;
      while ii <= numel (fmt_words)

        nxt_wrd = 0;
        ## Keep track of which words nr. every fmt_words{} is (starts) in.
        fwptr(ii) = iwrd;

        if (idr(ii))
          ## Regexp-like specifier. Search current + next words until satisfied
          ## sttc & endc keep track of columns to be concatenated later on
          sttc(ii) = jj = iwrd;
          ptrn = fmt_words{ii}(index (fmt_words{ii}, '['): end);
          while jj <= numel (words)
            [rs, re, ~, ~, ~, ~, nm] = regexp (words{jj}(iwrdp+1:end), ptrn);
            if (isempty (rs))
              ## Apparently an empty match; stop search & leave iwrdp untouched
              iwrd = jj;
              iwrdl = length (words{jj});
              jj = numel(words) + 1;
            else
              if (! length ([nm{:}]) == 0)
                ## Match before end of word
                ipos = find (! cellfun ("isempty", nm)) - 1;
                iwrdp += ipos(1);
                ## Provision for EOL: in that case bump {words} ptr
                if (words{jj}(iwrdp+1) == char(253)); nxt_wrd = 1; endif
                iwrd = jj;
                jj = numel (words) + 1;
              else
                ## Match includes at last char of words{jj}
                if (idw(ii))
                  ## Pattern contains whitespace/delims
                  ## => next word may contain more matches. Continue there
                  ++jj;
                  if (jj <= numel (words))
                    iwrdl = length (words{jj});
                    iwrdp = 0;
                  else
                    iwrd = jj;
                    nxt_wrd = 1;
                  endif
                else
                  ## Match should end at delimiter
                  iwrd = jj;
                  jj = numel (words) + 1;
                  nxt_wrd = 1;
                endif
              endif
            endif
          endwhile
          endc(ii) = min (iwrd, numel (words));

        elseif (idq(ii))
          ## double quote specifier (%q). Check if matching words contain delims
          ## sttc & endc keep track of columns to be concatenated later on
          sttc(ii) = jj = iwrd;
          ## Find position of first '"' (might occur after whitespace)
          ## Consume '"' & set iwrdp before first expected matching '"' position
          iwrdp += regexp (words{iwrd}(iwrdp+1:end), '[^ ]*"');
          ## Find word containing next '"'
          while jj <= numel (words)
            ## Find position of matching double quote
            if (iwrdp < length (words{jj}))
              qtepos = strfind (words{jj}(iwrdp+1:end), '"');
              if (! isempty (qtepos))
                ## Found. First dbl quote of next %q string may also be here
                iwrdp += qtepos(1);
                iwrd = jj;
                if (iwrdp == length (words{jj})); 
                  nxt_wrd = 1; 
                endif
                jj = numel (words) + 1;
              else
                ## Matching '"' must be in a next word.
                ++jj; iwrdp = 0;
                if (jj <= numel (words))
                  iwrdl = length (words{jj});
                endif
              endif
            else
              ## Matching '"' must be in a next word.
              ++jj; iwrdp = 0;
              if (jj <= numel (words))
                iwrdl = length (words{jj});
              endif
            endif
          endwhile
          endc(ii) = min (iwrd, numel (words));

        elseif (idf(ii))
          ## Literal expected
          if (isempty (strfind (fmt_words{ii}, words(iwrd))))
            ## Not found in current word; supposed to be in next word
            nxt_wrd = 1;
          else
            ## Found it in current word.  Subtract literal length
            iwrdp += length (fmt_words{ii});
            if (iwrdp > iwrdl)
              ## Parse error.  Literal extends beyond delimiter (word boundary)
              warning ("strread: literal '%s' (fmt spec # %d) does not match data", ...
                fmt_words{ii}, ii);
              ## Word assumed to be completely "used up". Next word
              nxt_wrd = 1;
            elseif (iwrdp == iwrdl)
              ## Word completely "used up". Next word
              nxt_wrd = 1;
            endif
          endif

        elseif (idg(ii))
          ## Fixed width specifier (%N or %*N): read just a part of word
          sw = regexp (fmt_words{ii}, '\d', "once");
          ew = regexp (fmt_words{ii}, '[nfudsq]') - 1;
          iwrdp += floor (str2double (fmt_words{ii}(sw:ew)));
          if (iwrdp > iwrdl)
            ## Match error. Field extends beyond word boundary.
            warning  ...
            ("strread: field width '%s' (fmt spec # %d) extends beyond actual word limit", ...
               fmt_words{ii}, ii);
            ## Assume word to be completely "used up".  Next word
            nxt_wrd = 1;
          elseif (iwrdp == iwrdl)
            ## Word completely "used up".  Next word
            nxt_wrd = 1;
          endif

        else
          ## A simple format conv. specifier. Either (1) uses rest of word, or
          ## (2) is squeezed between current iwrdp and next literal, or (3) uses
          ## next word. (3) is already taken care of.  So just check (1) & (2)
          if (ii < numel (fmt_words) && idf(ii+1))
            ## Next fmt_word is a literal...
            if (! index (words{iwrd}(iwrdp+1:end), fmt_words{ii+1}))
              ## ...but not found in current word => field uses rest of word
              nxt_wrd = 1;
            else
              ## ..or it IS found.  Add inferred width of current conversion field
              iwrdp += index (words{iwrd}(iwrdp+1:end), fmt_words{ii+1}) - 1;
            endif
          elseif (iwrdp < iwrdl)
            ## No bordering literal to the right => field occupies (rest of) word
            nxt_wrd = 1;
          endif

        endif

        if (nxt_wrd)
          ++iwrd; iwrdp = 0;
          if (iwrd > numel (words))
            ## Apparently EOF; assume incomplete row already at L.1 of data
            ii = numel (fmt_words);
          elseif (ii < numel (fmt_words) && iwrd <= numel (words))
            iwrdl = length (words{iwrd});
          endif
        endif

        ++ii;

      endwhile
      ## Done
      words_period = max (iwrd - 1, 1);
      num_lines = ceil (num_words / words_period);

      ## 2. Pad words array so that it can be reshaped
      num_words_padded = num_lines * words_period - num_words;
      if (num_words_padded)
        words = [words'; cell(num_words_padded, 1)];
      endif
      words = reshape (words, words_period, num_lines);

      ## 3. Concatenate columns contained within single %q or %[] specifiers
      ## Proceed backward (right to left) to avoid mixing up pointers
      icol = length (fmt_words);
      while (icol > 0)
        if (sttc(icol) > 0)
          ## For this fmt specifier some columns must be combined
          ncols = endc(icol) - sttc(icol) + 1;
          if (ncols > 1)
            ## Combine columns. Transpose as combining later on is columnwise
            tmpc = words(sttc(icol):endc(icol), :);
            ## For %q specifiers, insert space columns
            if (idq(icol))
              for ii=1:ncols-1
                jj = ncols - ii;
                tmpc = [tmpc(1:jj+1, :); tmpc(jj+1:end, :)];
                tmpc(jj+1, :) = " ";
              endfor
            endif
            mcols = size (tmpc, 1);
            ## Add split to row below
            ## FIXME: this assumes data doesn't contain char(255) 
            tmpc(mcols+1, :) = char(255);
            ## Assign combined columns to start column
            words(sttc(icol), :) = strsplit ([tmpc{1:end-1}], char (255));
            words(sttc(icol)+1:endc(icol), :) = [];
            ## Update fmt_words pointers to words
            if (icol < length (fmt_words))
              fwptr(icol+1:end) = fwptr(icol+1:end) - ncols + 1;
            endif
          endif
        endif
        --icol;
      endwhile

      ## 4. Do the column splitting on rectangular words array
      icol = 1; ii = 1;    # icol = current column, ii = current fmt_word
      while (ii <= num_words_per_line)

        ## Check if fmt_words(ii) contains a literal or fixed-width
        if ((idf(ii) || idg(ii)) && (rows(words) < num_words_per_line))
          if (idf(ii))
            s = strfind (words(icol, 1), fmt_words{ii});
            if (isempty (s{:}))
              error ("strread: literal '%s' not found in column %d", fmt_words{ii}, icol);
            endif
            s = s{:}(1);
            e = s(1) + length (fmt_words{ii}) - 1;
          endif
          if (! strcmp (fmt_words{ii}, words{icol, 1}))
            ## Column doesn't exactly match literal => split needed.  Insert a column
            words(icol+1:end+1, :) = words(icol:end, :);
            ## Watch out for empty cells
            jptr = find (! cellfun ("isempty", words(icol, :)));

            ## Distinguish leading or trailing literals
            if (! idg(ii) && ! isempty (s) && s(1) == 1)
              ## Leading literal.  Assign literal to icol, paste rest in icol + 1
              ## Apply only to those cells that do have something beyond literal
              jptr = find (cellfun ("length", words(icol+1, jptr), ...
                            "UniformOutput", false) > e(1));
              words(icol+1, :) = {""};
              words(icol+1, jptr) = cellfun ...
                (@(x) substr(x, e(1)+1, length(x)-e(1)), words(icol, jptr), ...
                "UniformOutput", false);
              words(icol, jptr) = fmt_words{ii};
              fwptr = [fwptr(1:ii) (++fwptr(ii+1:end))];

            else
              if (! idg(ii) && ! isempty (strfind (fmt_words{ii-1}, "%s")))
                ## Trailing literal.  If preceding format == '%s' this is an error
                warning ("Ambiguous '%s' specifier next to literal in column %d", icol);
              elseif (idg(ii))
                ## Current field = fixed width. Strip into icol, rest in icol+1
                sw = regexp (fmt_words{ii}, '\d', "once");
                ew = regexp (fmt_words{ii}, '[nfudsq]') - 1;
                wdth = floor (str2double (fmt_words{ii}(sw:ew)));
                words(icol+1, jptr) = cellfun (@(x) x(wdth+1:end),
                     words(icol,jptr), "UniformOutput", false);
                if (isempty ([words(icol+1, :){:}]))
                  ## Apparently split wasn't needed as turns out to cover
                  ## entire column. So delete column again
                  words(icol+1, :) = [];
                else
                  words(icol, jptr) = strtrunc (words(icol, jptr), wdth);
                  fwptr = [fwptr(1:ii) (++fwptr(ii+1:end))];
                endif
              else
                ## FIXME: this assumes char(254)/char(255) won't occur in input!
                clear wrds;
                wrds(1:2:2*numel (words(icol, jptr))) = ...
                     strrep (words(icol, jptr), fmt_words{ii}, ...
                     [char(255) char(254)]);
                wrds(2:2:2*numel (words(icol, jptr))-1) = char(255);
                wrds = strsplit ([wrds{:}], char(255));
                words(icol, jptr) = ...
                  wrds(find (cellfun ("isempty", strfind (wrds, char(254)))));
                wrds(find (cellfun ("isempty", strfind (wrds, char(254))))) ...
                   = char(255);
                words(icol+1, jptr) = strsplit (strrep ([wrds{2:end}], ...
                   char(254), fmt_words{ii}), char(255));
                ## Former trailing literal may now be leading for next specifier
                --ii;
                fwptr = [fwptr(1:ii) (++fwptr(ii+1:end))];
              endif
            endif
          endif

        elseif (idr(ii) && ii < num_words_per_line)
          ## Leading %[..] type specifier. Split into two columns. This split
          ## is simply unavoidable as we don't know what is left to follow %[]
          ## 1. Amend pattern
          ptrn = ['^' fmt_words{ii}(index (fmt_words{ii}, '['): end) '*'];
          ## 2. Extract leading column and remove optional trailing char(253)
          tmpwrds = words(icol, :);
          words(icol, :) = regexp (words(icol, :), ptrn, 'match', 'once');
          if (words{icol, 1}(end) == char(253))
            words(icol, :) = strrep (words(icol, :), char(253), '');
          endif
          ## 3. Extract trailing column
          tmpwrds = regexprep (tmpwrds, ptrn, '');
          ## 4. This split may not have been needed; delete temp column if empty
          if (isempty ([tmpwrds{:}]))
            clear tmpwrds;
          else
            ## 5. Insert a column and add tmpwrds after present column 
            words(icol+2:end+1, :) = words(icol+1:end, :);
            words(icol+1, :) = tmpwrds;
            fwptr = [fwptr(1:ii) (++fwptr(ii+1:end))];
          endif

        elseif (idr(ii))
          ## Just check for trailing char(253) and if present remove it
          if (words{icol, 1}(end) == char(253))
            words(icol, :) = strrep (words(icol, :), char(253), '');
          endif

        elseif (idq(ii))
          ## In all cases remove left '"' 
          words(icol, :) = regexprep (words(icol, :), '^"', "");
          if  (ii < num_words_per_line)
            ## Check if a column to the right must be split off
            if (fwptr(icol) == fwptr(icol+1))
              ## Replace only first occurrence of '"', next %q may be in same col
              words(icol, :) = regexprep (words(icol, :), '"', char(255), "once");
              words(icol+1:end+1, :) = words(icol:end, :);
              ptrn = ["^.*" char(255)];
              ## Right column
              words(icol+1, :) = regexprep (words(icol+1, :), ptrn, "", "once");
              ## Left column
              words(icol, :) = regexprep (words(icol, :), [char(255) ".*$"], "");
              fwptr = [fwptr(1:ii) (++fwptr(ii+1:end))];
            endif
          endif
          ## Replace matching occurrence of '"'
          words(icol, :) = regexprep (words(icol, :), '"', "", "once");

        else
          ## Conv. specifier.  Peek if next fmt_word needs split from current column
          if (ii < num_words_per_line)
            if (fwptr(ii) == fwptr(ii+1))
              --icol;
            endif
          endif
        endif
        ## Next fmt_word, next column
        ++ii; ++icol;
      endwhile

      ## Done.  Reshape words back into 1 long vector and strip padded empty words
      words = reshape (words, 1, numel (words))(1 : end-num_words_padded);

    catch
      warning ("strread: unable to parse text or file with given format string");
      return;

    end_try_catch
  endif

  ## For each specifier, process corresponding column
  k = 1;
  for m = 1:num_words_per_line
    try
      if (format_repeat_count < 0)
        data = words(m:num_words_per_line:end);
      elseif (format_repeat_count == 0)
        data = {};
      else
        lastline = ...
          min (num_words_per_line * format_repeat_count + m - 1, numel (words));
        data = words(m:num_words_per_line:lastline);
      endif

      ## Map to format
      ## FIXME - add support for formats like "<%s>", "%[a-zA-Z]"
      ##         Someone with regexp experience is needed.
      switch fmt_words{m}(1:min (2, length (fmt_words{m})))
        case {"%s", "%[", "%q"}
          ## for %[] & %q, column has been singled out above; return as string
          if (pad_out)
            data(end+1:num_lines) = {""};
          endif
          varargout{k} = data';
          k++;
        case {"%d", "%u", "%f", "%n"}
          n = cellfun ("isempty", data);
          ### FIXME - erroneously formatted data lead to NaN, not an error
          data = str2double (data);
          if (! isempty (regexp (fmt_words{m}, "%[du]")))
            ## Cast to integer
            ## FIXME: NaNs will be transformed into zeros
            switch fmt_words{m}
              case '%d8'
                data = int8 (data);
              case '%d16'
                  data = int16 (data);
              case '%d64'
                  data = int64 (data);
              case '%u8'
                data = uint8 (data);
              case '%u16'
                  data = uint16 (data);
              case {'%u32', '%u'}
                  data = uint32 (data);
              case '%u64'
                  data = uint64 (data);
              otherwise
                data = int32 (data);
            endswitch
          elseif (regexp (fmt_words{m}, '%f32'))
            data = single (data);
          endif
          data(n) = numeric_fill_value;
          if (pad_out)
            data(end+1:num_lines) = numeric_fill_value;
          endif
          varargout{k} = data.';
          k++;
        case {"%0", "%1", "%2", "%3", "%4", "%5", "%6", "%7", "%8", "%9"}
          sw = regexp (fmt_words{m}, '\d', "once");
          ew = regexp (fmt_words{m}, '[nfudsq]') - 1;
          nfmt = strsplit (fmt_words{m}(2:ew), ".");
          swidth = str2double (nfmt{1});
          switch fmt_words{m}(ew+1)
            case {"d", "u", "f", "n"}
              n = cellfun ("isempty", data);
              ## FIXME - erroneously formatted data lead to NaN, not an error
              ##         => ReturnOnError can't be implemented for numeric data
              data = str2double (data);
              if (! isempty (regexp (fmt_words{m}, "f32")))
                data = single (data);
              endif
              if (pad_out)
                data(end+1:num_lines) = numeric_fill_value;
              endif
              if (! isempty (regexp (fmt_words{m}, "[du]")))
                ## Cast to integer
                ## FIXME: NaNs will be transformed into zeros
                s = regexp (fmt_words{m}, "[du]", "start");
                switch fmt_words{m}(s : end)
                  case "d8"
                    data = int8 (data);
                  case "d16"
                    data = int16 (data);
                  case "d64"
                    data = int64 (data);
                  case "u8"
                    data = uint8 (data);
                  case "u16"
                    data = uint16 (data);
                  case {"u32", "u"}
                    data = uint32 (data);
                  case "u64"
                    data = uint64 (data);
                  otherwise
                    data = int32 (data);
                endswitch
              else
                if (! isempty (strfind (fmt_words{m}, "f32")))
                  data = single (data);
                  if (numel (nfmt) > 1)
                    sprec = str2double (nfmt{2});
                    data = 10^-sprec * single (round (10^sprec * data));
                  endif
                else
                  if (numel (nfmt) > 1)
                    sprec = str2double (nfmt{2});
                    data = 10^-sprec * double (round (10^sprec * data));
                  endif
                endif
              endif
              data(n) = numeric_fill_value;
              varargout{k} = data.';
              k++;
            case "s"
              if (pad_out)
                data(end+1:num_lines) = {""};
              endif
              varargout{k} = strtrunc (data, swidth)';
              k++;
            otherwise
          endswitch
        case {"%*", "%*s"}
          ## skip the word
        otherwise
          ## Ensure descriptive content is consistent.
          ## Test made a bit lax to accomodate for incomplete last lines
          n = find (! cellfun ("isempty", data));
          if (numel (unique (data(n))) > 1
              || ! strcmpi (unique (data), fmt_words{m}))
            error ("strread: FORMAT does not match data");
          endif
      endswitch
    catch
      ## As strread processes columnwise, ML-compatible error processing
      ## (row after row) is not feasible. In addition Octave sets unrecognizable
      ## numbers to NaN w/o error.  But maybe Octave is better in this respect.
      if (err_action)
        ## Just try the next column where ML bails out
      else
        rethrow (lasterror);
      endif
    end_try_catch
  endfor

endfunction

function out = split_by (text, sep, mult_dlms_s1, eol_char)

  ## Check & if needed, process MultipleDelimsAsOne parameter
  if (mult_dlms_s1)
    mult_dlms_s1 = true;
    ## FIXME: Should re-implement strsplit() function here in order
    ## to avoid strrep on megabytes of data.
    ## If \n is in sep collection we need to enclose it in text
    ## to avoid it being included in consecutive delim series
    enchr = ' ';
    ## However watch out if eol_char is also in delimiters
    if (index (sep, eol_char)); enchr = char(255); endif
    text = strrep (text, eol_char, [enchr eol_char enchr]);
  else
    mult_dlms_s1 = false;
  endif

  ## Split text string along delimiters
  out = strsplit (text, sep, mult_dlms_s1);
  if (index (sep, eol_char)); out = strrep (out, char(255), ''); endif
  ## In case of trailing delimiter, strip stray last empty word
  if (!isempty (out) && any (sep == text(end)))
    out(end) = [];
  endif

  ## Empty cells converted to empty cellstrings.
  out(cellfun ("isempty", out)) = {""};

endfunction


%!test
%! [a, b] = strread ("1 2", "%f%f");
%! assert (a, 1);
%! assert (b, 2);

%!test
%! str = "";
%! a = rand (10, 1);
%! b = char (randi ([65, 85], 10, 1));
%! for k = 1:10
%!   str = sprintf ("%s %.6f %s\n", str, a(k), b(k));
%! endfor
%! [aa, bb] = strread (str, "%f %s");
%! assert (a, aa, 1e-6);
%! assert (cellstr (b), bb);

%!test
%! str = "";
%! a = rand (10, 1);
%! b = char (randi ([65, 85], 10, 1));
%! for k = 1:10
%!   str = sprintf ("%s %.6f %s\n", str, a(k), b(k));
%! endfor
%! aa = strread (str, "%f %*s");
%! assert (a, aa, 1e-6);

%!test
%! str = sprintf ("/* this is\nacomment*/ 1 2 3");
%! a = strread (str, "%f", "commentstyle", "c");
%! assert (a, [1; 2; 3]);

%!test
%! str = "# comment\n# comment\n1 2 3";
%! [a, b] = strread (str, "%n %s", "commentstyle", "shell", "endofline", "\n");
%! assert (a, [1; 3]);
%! assert (b, {"2"});

%!test
%! str = sprintf ("Tom 100 miles/hr\nDick 90 miles/hr\nHarry 80 miles/hr");
%! fmt = "%s %f miles/hr";
%! c = cell (1, 2);
%! [c{:}] = strread (str, fmt);
%! assert (c{1}, {"Tom"; "Dick"; "Harry"})
%! assert (c{2}, [100; 90; 80])

%!test
%! a = strread ("a b c, d e, , f", "%s", "delimiter", ",");
%! assert (a, {"a b c"; "d e"; ""; "f"});

%!test
%! # Bug #33536
%! [a, b, c] = strread ("1,,2", "%s%s%s", "delimiter", ",");
%! assert (a{1}, "1");
%! assert (b{1}, "");
%! assert (c{1}, "2");

%!test
%! # Bug #33536
%! a = strread ("[SomeText]", "[%s", "delimiter", "]");
%! assert (a{1}, "SomeText");

%!test
%! dat = "Data file.\r\n=  =  =  =  =\r\nCOMPANY    : <Company name>\r\n";
%! a = strread (dat, "%s", "delimiter", "\n", "whitespace", "", "endofline", "\r\n");
%! assert (a{2}, "=  =  =  =  =");
%! assert (double (a{3}(end-5:end)), [32 110 97 109 101 62]);

%!test
%! [a, b, c, d] = strread ("1,2,3,,5,6", "%d%f%d%f", "delimiter", ",");
%! assert (c, int32 (3));
%! assert (d, NaN);

%!test
%! [a, b, c, d] = strread ("1,2,3,,5,6\n", "%d%d%f%d", "delimiter", ",");
%! assert (c, [3; NaN]);
%! assert (d, int32 ([0; 0]));

%!test
%! # Default format (= %f)
%1 [a, b, c] = strread ("0.12 0.234 0.3567");
%1 assert (a, 0.12);
%1 assert (b, 0.234);
%1 assert (c, 0.3567);

%!test
%! [a, b] = strread ("0.41 8.24 3.57 6.24 9.27", "%f%f", 2, "delimiter", " ");
%1 assert (a, [0.41; 3.57]);

%!test
%! # TreatAsEmpty
%! [a, b, c, d] = strread ("1,2,3,NN,5,6\n", "%d%d%d%f", "delimiter", ",", "TreatAsEmpty", "NN");
%! assert (c, int32 ([3; 0]));
%! assert (d, [NaN; NaN]);

%!test
%! # No delimiters at all besides EOL.  Plain reading numbers & strings
%! str = "Text1Text2Text\nText398Text4Text\nText57Text";
%! [a, b] = strread (str, "Text%dText%1sText");
%! assert (a, int32 ([1; 398; 57]));
%! assert (b(1:2), {"2"; "4"});
%! assert (isempty (b{3}), true);

%% MultipleDelimsAsOne
%!test
%! str = "11, 12, 13,, 15\n21,, 23, 24, 25\n,, 33, 34, 35";
%! [a b c d] = strread (str, "%f %f %f %f", "delimiter", ",", "multipledelimsasone", 1, "endofline", "\n");
%! assert (a', [11, 21, NaN]);
%! assert (b', [12, 23, 33]);
%! assert (c', [13, 24, 34]);
%! assert (d', [15, 25, 35]);

%% delimiter as sq_string and dq_string
%!test
%! assert (strread ("1\n2\n3", "%d", "delimiter", "\n"),
%!         strread ("1\n2\n3", "%d", "delimiter", '\n'))

%% whitespace as sq_string and dq_string
%!test
%! assert (strread ("1\b2\r3\b4\t5", "%d", "whitespace", "\b\r\n\t"),
%!         strread ("1\b2\r3\b4\t5", "%d", "whitespace", '\b\r\n\t'))

%!test
%! str =  "0.31 0.86 0.94\n 0.60 0.72 0.87";
%! fmt = "%f %f %f";
%! args = {"delimiter", " ", "endofline", "\n", "whitespace", " "};
%! [a, b, c] = strread (str, fmt, args {:});
%! assert (a, [0.31; 0.60], 0.01)
%! assert (b, [0.86; 0.72], 0.01)
%! assert (c, [0.94; 0.87], 0.01)

%!test
%! str =  "0.31,0.86,0.94\n0.60,0.72,0.87";
%! fmt = "%f %f %f";
%! args = {"delimiter", ",", "endofline", "\n", "whitespace", " "};
%! [a, b, c] = strread (str, fmt, args {:});
%! assert (a, [0.31; 0.60], 0.01)
%! assert (b, [0.86; 0.72], 0.01)
%! assert (c, [0.94; 0.87], 0.01)

%!test
%! str =  "0.31 0.86 0.94\n 0.60 0.72 0.87";
%! fmt = "%f %f %f";
%! args = {"delimiter", ",", "endofline", "\n", "whitespace", " "};
%! [a, b, c] = strread (str, fmt, args {:});
%! assert (a, [0.31; 0.60], 0.01)
%! assert (b, [0.86; 0.72], 0.01)
%! assert (c, [0.94; 0.87], 0.01)

%!test
%! str =  "0.31, 0.86, 0.94\n 0.60, 0.72, 0.87";
%! fmt = "%f %f %f";
%! args = {"delimiter", ",", "endofline", "\n", "whitespace", " "};
%! [a, b, c] = strread (str, fmt, args {:});
%! assert (a, [0.31; 0.60], 0.01)
%! assert (b, [0.86; 0.72], 0.01)
%! assert (c, [0.94; 0.87], 0.01)

%!test
%! [a, b] = strread (["Empty 1" char(10)], "Empty%s %f");
%! assert (a{1}, '1');
%! assert (b, NaN);

%!test
%! [a, b] = strread (["Empty" char(10)], "Empty%f %f");
%! assert (a, NaN);
%! assert (b, NaN);

%!test
%! # Bug #35999
%! [a, b, c] = strread ("", "%f");
%! assert (isempty (a));
%! assert (isempty (b));
%! assert (isempty (c));

%% Bit width, fixed width conv. specifiers
%!test
%! str2 = "123456789012345 ";
%! str2 = [str2 str2 str2 str2 str2 str2 str2 str2];
%! str2 = [str2 " 123456789.01234 1234567890.1234 12345.678901234 12345.678901234"];
%! pttrn = "%3u8%*s %5u16%*s %10u32%*s %15u64 %3d8%*s %5d16%*s %10d32%*s %15d64 %9f32%*s %14f64%*s %10.2f32%*s %12.2f64%*s";
%! [a b c d e f g h i j k l] = strread (str2, pttrn, "delimiter", " ");
%! assert (a, uint8 (123));
%! assert (b, uint16 (12345));
%! assert (c, uint32 (1234567890));
%! assert (d, uint64 (123456789012345));
%! assert (e, int8 (123));
%! assert (f, int16 (12345));
%! assert (g, int32 (1234567890));
%! assert (h, int64 (123456789012345));
%! assert (i, single (123456789), 1e-12);
%! assert (j, double (1234567890.123), 1e-15);
%! assert (k, single (12345.68), 1e-5);
%! assert (l, double (12345.68), 1e-15);

%% %[] tests
%% Plain [..] and *[..]
%!test
%! ar = "abcdefguvwxAny\nacegxyzTrailing\nJunk";
%! [a, b] = strread (ar, "%[abcdefg] %*[uvwxyz] %s");
%! assert (a, {"abcdefg"; "aceg"; ""});
%! assert (b, {"Any"; "Trailing"; "Junk"});

%!test
%! assert (strread ("A2 B2 C3", "%*[ABC]%d", 3), int32([2; 2; 3]));

%% [^..] and *[^..]
%!test
%! br = "abcdefguvwx1Any\nacegxyz2Trailing\n3Junk";
%! [a, b] = strread (br, "%[abcdefg] %*[^0123456789] %s");
%! assert (a, {"abcdefg"; "aceg"; ""});
%! assert (b, {"1Any"; "2Trailing"; "3Junk"});

%% [..] and [^..] crossing delimiters
%!test
%! cr = "ab cd efguv wx1Any\na ce gx yz2Trailing\n   3Junk";
%! [a, b] = strread (cr, "%[ abcdefg] %*[^0123456789] %s", "delimiter", " \n", "whitespace", "");
%! assert (a, {"abcdefg"; "aceg"; ""});
%! assert (b, {"1Any"; "2Trailing"; "3Junk"});

%% Bug #36464
%!test
%! assert (strread ('1 2 3 4 5 6', "%*n%n%*[^\n]"), 2);

%% test %[]] and %[^]]
%!test
%! assert (strread ('345]', "%*[123456]%[]]"){1}, "]");
%! assert (strread ('345]', "%*[^]]%s"){1}, "]");

%% Textscan example
%!test
%! st4 = {"1st string  123.45 MoreText 1953-05-14 60 82+i15 abcdefg", "\n";
%!        "2nd string  678.90 LessText 1961-02-01 52 53-i03 aceg",    "\n";
%!        "3rd string  246.08 SameText 1973-12-13 40 12+i66 bdfh",    "\n";
%!        "4th string  135.79 NoText 1982-09-27 29 13+i81 iiiiii",    "\n"};
%! C = textscan ([st4'{:}], "%s %*[^\n]");
%! D = textscan ([st4'{:}], "%s string %n %*[^\n]");
%! E = textscan ([st4'{:}], "%s string %n %[^T]Text %*[^\n]");
%! assert (C{1}, {"1st"; "2nd"; "3rd"; "4th"});
%! assert (D{1}, {"1st"; "2nd"; "3rd"; "4th"});
%! assert (D{2}, [123.45; 678.90; 246.08; 135.79], 1e-10);
%! assert (E{1}, {"1st"; "2nd"; "3rd"; "4th"});
%! assert (E{2}, [123.45; 678.90; 246.08; 135.79], 1e-10);
%! assert (E{3}, {"More"; "Less"; "Same"; "No"});

%% %q test, simple case
%! str = 'aaa "Notalongtext" bbb "AnotherShortText"';
%! [a, b] = strread (str, "%s %q");
%! assert (a, {"aaa"; "bbb"});
%! assert (b, {"Notalongtext"; "AnotherShortText"});

%% %q, with consecutive white space and delims between double quotes
%!test
%! str = ['25 "ab' "\t" 'cd" 34.2' "\n" '8 "dggs\\ ' "\n" 'fdf" 101.3'];
%! [a, b, c]  = strread (str, "%d %q %f");
%! assert (a, int32([25; 8]));
%! assert (b, {"ab cd"; 'dggs\\ fdf'});   ## Single quotes, double backslash 
%! assert (c, [34.2; 101.3], 1e-15);

%% %q test, consecutive %q words
%!test
%! str = '" a a""bb" " c c""dd"';
%! [a, b] = strread (str, "%q%q");
%! assert (a, {" a a"; " c c"});
%! assert (b, {"bb"; "dd"});

%% %q with delimiters (whitespace) contained within column w immediately adjacent col.
%!test
%! str =          '1 Harry "17  Station St, Milw"02/05/1973';
%! str = [str "\n" '2 Doug "1A  Through Rd, Merl"22/08/1981' "\n"];
%! str = [str "\n" '3 John "581 Main St,    Lard"03/01/1979' "\n"];
%! [a, b, c, d] = strread (str, "%d %s %q%s");
%! assert (a', int32([1 2 3]));
%1 assert (b, {"Harry"; "Doug"; "John"});
%! assert (c, {"17 Station St, Milw"; "1A Through Rd, Merl"; "581 Main St, Lard"});
%! assert (d, {"02/05/1973"; "22/08/1981"; "03/01/1979"});

%% Unsupported format specifiers
%!test
%!error <format specifiers are not supported> strread ("a", "%c")
%!error <format specifiers are not supported> strread ("a", "%*c %d")
